<?php

namespace Drupal\viz_builder\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Asset\AssetResolver;
use Drupal\Core\Asset\AttachedAssets;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Render\Renderer;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Extension\ModuleHandler;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\DependencyInjection\ContainerInterface;

class ApiController extends ControllerBase {

  /**
   * @var \Drupal\Core\Asset\AssetResolver
   */
  protected $assetResolver;

  /**
   * @var \Drupal\Core\Render\Renderer
   */
  protected $renderer;

  /**
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected $fileUrlGenerator;

  public function __construct(
    AssetResolver $asset_resolver,
    FileUrlGeneratorInterface $file_url_generator,
    Renderer $renderer
  ) {
    $this->assetResolver = $asset_resolver;
    $this->fileUrlGenerator = $file_url_generator;
    $this->renderer = $renderer;
  }

  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('asset.resolver'),
      $container->get('file_url_generator'),
      $container->get('renderer')
    );
  }

  public function loadThemeAssets() {
    $theme = \Drupal::theme()->getActiveTheme();

    $build = ['#attached' => ['html_response_attachment_placeholders' => []]];
    \Drupal::service('main_content_renderer.html')->invokePageAttachmentHooks($build);
    $assets = AttachedAssets::createFromRenderArray($build);
    $libs = array_merge([
      'core/drupal.ajax'
    ], $theme->getLibraries());
    $assets->setLibraries($libs);
    $rendered_css = [];    
    foreach($this->assetResolver->getCssAssets($assets, FALSE) as $key => &$item) {
      $item['attributes']['data-key'] = $key;
      $renderable = \Drupal::service('asset.css.collection_renderer')->render([$item]);
      $rendered_css[] = [
        'key' => $key,
        'content' => $this->renderer->renderPlain($renderable)
      ];
    }

    $jsRenderer = \Drupal::service('asset.js.collection_renderer');
    $js_assets = $this->assetResolver->getJsAssets($assets, FALSE);
    $rendered_js = [];
    $settings = new \stdClass();
    foreach($js_assets[0] as $key => &$item) {
      $renderable = $jsRenderer->render([$item]);
      $rendered_js[] = [
        'position' => 'head',
        'key' => $key,
        'src' => $renderable[0]['#attributes']['src'],
      ];
    }
    foreach($js_assets[1] as $key => &$item) {
      if ($key == 'drupalSettings') {
        $settings = $item['data'];
        continue;
      };
      $renderable = $jsRenderer->render([$item]);
      $rendered_js[] = [
        'position' => 'body',
        'key' => $key,
        'src' => $renderable[0]['#attributes']['src'],
      ];
    }

    return new JsonResponse([
      'raw' => $assets->getLibraries(),
      'theme' => $theme->getName(),
      // 'css' => $css,
      'settings' => $settings,
      'css' => $rendered_css,
      'js' => $rendered_js,
    ]);
  }

  public function render () {
    $request = \Drupal::request();
    $payload = Json::decode($request->getContent());
    $template = $payload['template'] ?? '';
    $context = $payload['context'] ?: [];
    $context = $this->moduleHandler()->invokeAll('twig_formatter_context', [$context, NULL]);
    $build = [
      '#type' => 'inline_template',
      '#template' => $template,
      '#context' => array_merge(
        [],
        $context
      ),
      '#attached' => []
    ];
    // need set master route to editing page
    // \Drupal::service('main_content_renderer.html')->invokePageAttachmentHooks($build);
    $html = $this->renderer->renderRoot($build);
    $attached = AttachedAssets::createFromRenderArray($build);

    $css_assets = $this->assetResolver->getCssAssets($attached, false);
    $js_assets = $this->assetResolver->getJsAssets($attached, false);
    $rendered_css = [];
    $rendered_js = [];
    $settings = new \stdClass();
    
    foreach($css_assets as $key => &$item) {
      $item['attributes']['data-key'] = $key;
      $renderable = \Drupal::service('asset.css.collection_renderer')->render([$item]);
      $rendered_css[] = [
        'key' => $key,
        'content' => $this->renderer->renderPlain($renderable)
      ];
    }
    foreach($js_assets[1] as $key => &$item) {
      if ($key == 'drupalSettings') {
        $settings = $item['data'];
        continue;
      };
      $item['attributes']['data-key'] = $key;
      $renderable = \Drupal::service('asset.js.collection_renderer')->render([$item]);
      $rendered_js[] = [
        // 'key' => $key,
        'src' => $renderable[0]['#attributes']['src'],
      ];
    }

    return new JsonResponse([
      'html' => $html, 
      'settings' => $settings, 
      'js' => $rendered_js, 
      'css' => $rendered_css
    ]);
  }

  public function componentList() {
    $comps = [];
    /** @var \Drupal\sdc\ComponentPluginManager $component_manager */
    $component_manager = \Drupal::service('plugin.manager.sdc');
    $components = $component_manager->getAllComponents();

    foreach($components as $component) {
      $meta = $component->metadata;
      $thumbnail = $meta->getThumbnailPath();
      $id = $component->getPluginId();
      if (strpos($id, 'cl_devel') > -1) continue;
      $plugin_def = $component->getPluginDefinition();
      $comps[] = [
        'name' => $id,
        'label' => $meta->name,
        'thumbnail' => $thumbnail ? $this->fileUrlGenerator->generateAbsoluteString($thumbnail) : '',
        'group' => $meta->group,
        // 'provider' => $component->getPluginId(),
        // 'twigName' => $meta->module,
        'schema' => @$plugin_def['props'],
        'uiSchema' => @$plugin_def['ui'],
        'default' => @$plugin_def['default'],
      ];
    }

    return new JsonResponse($comps);
  }

  public function componentDetail(string $component_id) {
    /** @var \Drupal\sdc\ComponentPluginManager $component_manager */
    $component_manager = \Drupal::service('plugin.manager.sdc');
    $component = $component_manager->find($component_id);
    $pluginDef = $component->getPluginDefinition();
    $data = [
      'name' => $component->metadata->machineName,
      'label' => $component->metadata->name,
      'group' => $component->metadata->group,
      // 'shortCode' => $component->
      // 'meta' => $component->metadata,
      'schema' => $pluginDef['props']
    ];

    return new JsonResponse($data);
  }

  public function blockList() {
    $definitions = \Drupal::service('plugin.manager.block')->getDefinitions();
    $blocks = [];
    foreach($definitions as $id => $definition) {
      if ($id == 'broken') continue;
      if (str_starts_with($id, 'field_block:') || str_starts_with($id, 'extra_field_block:')) continue;
      $blocks[] = [
        'id' => $id, 
        'label' => (string) $definition['admin_label']
      ];
    }

    return new JsonResponse($blocks);
  }
}
