<?php

namespace Drupal\viz_builder\Form;

use Drupal\Core\Config\Entity\ConfigEntityInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\FieldConfigInterface;
use Drupal\Core\Form\ConfirmFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;

class FieldConfirmForm extends ConfirmFormBase {
  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Entity\EntityDisplayRepository
   */
  protected $entityDisplayRepository;

  public function __construct() {
    $this->entityTypeManager = \Drupal::entityTypeManager();
    $this->entityDisplayRepository = \Drupal::service("entity_display.repository");

  }

  protected $field_map = [
    'twig_body' => [
      'lable' => 'Twig body',
      'type' => 'text_long',
    ],
    'builder_data' => [
      'lable' => 'Builder data',
      'type' => 'json_native',
    ],
  ];

  protected function getEntityBundleType() {
    $params = \Drupal::routeMatch()->getParameters();
    $entity_type_id = $params->get('entity_type_id');
    $bundle_id = $params->get('bundle_id');

    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    /** @var \Drupal\Core\Config\Entity\ConfigEntityInterface */
    $bundle_type = $this->entityTypeManager->getStorage($entity_type->getBundleEntityType())->load($bundle_id);
    return $bundle_type;
  }

  protected function isEnabled() {
    return $this->getEntityBundleType()->getThirdPartySetting('viz_builder', 'builder_enabled', FALSE);
  }

  protected function setEnabled($enabled = TRUE) {
    $this->getEntityBundleType()->setThirdPartySetting('viz_builder', 'builder_enabled', $enabled)->save();
  }

  public function getFormId() {
    return 'viz_builder_field_confirm_form';
  }

  public function getQuestion() {
    if ($this->isEnabled()) {
      return t('Disabling builder');
    } else {
      return t('Enabling builder');
    }
  }

  public function getCancelUrl() {
    return new Url('viz_builder.entity_config');
  }

  public function getDescription() {
    if ($this->isEnabled()) {
      return t('All builder content will be purged.<br/> <b>This action cannot be undone!</b>');
    }
    return t('Do you want to continue with this operation?');
  }

  public function getConfirmText() {
    return t('Continue');
  }

  public function submitForm(array &$form, FormStateInterface $form_state) {
    $params = \Drupal::routeMatch()->getParameters();
    $entity_id = $params->get('entity_type_id');
    $bundle_id = $params->get('bundle_id');

    if (!$this->isEnabled()) {
      foreach ($this->field_map as $machine_name => $field) {
        $storage_config = \Drupal\field\Entity\FieldStorageConfig::loadByName($entity_id, $machine_name);
        if (!$storage_config) {
          $storage_config = \Drupal\field\Entity\FieldStorageConfig::create([
            'field_name' => $machine_name,
            'entity_type' => $entity_id,
            'type' => $field['type'],
            'locked' => TRUE,
          ]);
          $storage_config->save();
        }
        $field_config = \Drupal\field\Entity\FieldConfig::loadByName($entity_id, $bundle_id, $machine_name);
        if (!$field_config) {
          $field_config = \Drupal\field\Entity\FieldConfig::create([
            'field_name' => $machine_name,
            'entity_type' => $entity_id,
            'bundle' => $bundle_id,
            'label' => $field['lable'],
          ]);
          $field_config->save();
        }
      }

      $jsonapi_resource_config = $this->getJsonapiResourceConfig($entity_id, $bundle_id);
      $resouce_fields = $jsonapi_resource_config->get('resourceFields');
      $resouce_fields['builder_data'] = [
        'fieldName' => 'builder_data',
        'publicName' => 'builder_data',
        'enhancer' => ['id' => 'json'],
        'disabled' => FALSE,
      ];
      $jsonapi_resource_config->set('resourceFields', $resouce_fields);
      $jsonapi_resource_config->save();

      $this->setEntityDisplay($entity_id, $bundle_id);

      $this->setEnabled();
      \Drupal::messenger()->addMessage(t('Builder enabled'));
      $form_state->setRedirectUrl($this->getCancelUrl());
    } else {
      foreach ($this->field_map as $machine_name => $field) {
        $field_config = \Drupal\field\Entity\FieldConfig::loadByName($entity_id, $bundle_id, $machine_name);
        if ($field_config) {
          $field_config->delete();
        }
      }
      $this->setEnabled(FALSE);
      \Drupal::messenger()->addMessage(t('Builder disabled'));
      $form_state->setRedirectUrl($this->getCancelUrl());
    }

  }

  /**
   * @return \Drupal\jsonapi_extras\Entity\JsonapiResourceConfig|null
   */
  protected function getJsonapiResourceConfig($entity_id, $bundle_id) {
    $storage = $this->entityTypeManager->getStorage('jsonapi_resource_config');

    if ($existing_config = $storage->load("$entity_id--$bundle_id")) {
      return $existing_config;
    }

    $entity_type = $this->entityTypeManager->getDefinition($entity_id);
    $fields = $this->getAllFieldNames($entity_type, $bundle_id);
    if (!empty($fields)) {
      return $storage->create([
        'id' => "$entity_id--$bundle_id",
        'disabled' => FALSE,
        'path' => "$entity_id/$bundle_id",
        'resourceType' => "$entity_id--$bundle_id",
        'resourceFields' => $fields,
      ]);
    }

    return NULL;
  }

  /**
   * Gets all field names for a given entity type and bundle.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   *   The entity type for which to get all field names.
   * @param string $bundle
   *   The bundle for which to get all field names.
   *
   * @return array
   *   All field definitions.
   */
  protected function getAllFieldNames(EntityTypeInterface $entity_type, $bundle_id) {
    $fields = [];

    if (is_a($entity_type->getClass(), FieldableEntityInterface::class, TRUE)) {
      /** @var \Drupal\Core\Entity\EntityFieldManager $entity_field_manager */
      $entity_field_manager = \Drupal::service("entity_field.manager");
      $field_definitions = $entity_field_manager->getFieldDefinitions(
        $entity_type->id(),
        $bundle_id
      );

      foreach ($field_definitions as $field_name => $field_definition) {
        $fields[$field_name] = [
          'fieldName' => $field_name,
          'publicName' => $field_name,
          'enhancer' => ['id' => ''],
          'disabled' => FALSE,
        ];
      }
    }

    return $fields;
  }

  protected function setEntityDisplay($entity_type_id, $bundle_id, $mode = 'default') {
    $view_display = $this->entityDisplayRepository->getViewDisplay($entity_type_id, $bundle_id, $mode);
    if ($view_display->status()) {
      $view_display->setComponent('twig_body', [
        'type' => 'twig_formatter',
        'label' => 'hidden',
        'settings' => [],
        'third_party_settings' => [],
        'weight' => 0,
      ]);
      $view_display->save();
    }
  }
}
