<?php

namespace Drupal\viz_builder\Plugin\jsonapi\FieldEnhancer;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\jsonapi_extras\Plugin\ResourceFieldEnhancerBase;
use Drupal\serialization\Normalizer\CacheableNormalizerInterface;
use Shaper\Util\Context;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Perform additional manipulations to timestamp fields.
 *
 * @ResourceFieldEnhancer(
 *   id = "file_url",
 *   label = @Translation("File Url (Image/file field)"),
 *   description = @Translation("Adds links for images/files")
 * )
 */
class FileUrl extends ResourceFieldEnhancerBase implements ContainerFactoryPluginInterface {

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $entity_type_manager = $container->get('entity_type.manager');
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $entity_type_manager
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function doUndoTransform($data, Context $context) {
    $config = $this->getConfiguration();
    $uuid_key = $this->entityTypeManager->getDefinition('file')->getKey('uuid');
    $entities = $this->entityTypeManager->getStorage('file')->loadByProperties([$uuid_key => $data['id']]);
    /** @var \Drupal\file\FileInterface $entity */
    $entity = reset($entities);
    // If the entity cannot be loaded or it's not an image, do not enhance it.
    if (!$entity) {
      return $data;
    }
    /** @var \Drupal\Core\Cache\CacheableMetadata $cacheableMetadata */
    $cacheableMetadata = $context->offsetGet(CacheableNormalizerInterface::SERIALIZATION_CONTEXT_CACHEABILITY)
      ->merge(CacheableMetadata::createFromObject($entity));
    /** @var \Drupal\file\Entity\File $entity */
    // If the entity is not viewable.
    $access = $entity->access('view', NULL, TRUE);
    if (!$access->isAllowed()) {
      return $data;
    }
    $meta = [
      'uri' => $entity->getFileUri(),
      'url' => $entity->createFileUrl(!$config['absolute']),
    ];
    // Must explicitly re-set because CacheableMetadata::merge() clones.
    $context->offsetSet(
      CacheableNormalizerInterface::SERIALIZATION_CONTEXT_CACHEABILITY,
      $cacheableMetadata
    );
    return array_merge_recursive($data, ['meta' => $meta]);
  }

  /**
   * Gets the list of image style IDs for the current field.
   *
   * Intersects the field enhancer configuration with the consumer's config to
   * calculate the list of image styles to apply to the field.
   *
   * @return string[]
   *   The list of IDs.
   */
  protected function imageStylesForField() {
    $configuration = $this->getConfiguration();
    $image_style_ids = $configuration['consumer_image_style_ids'];
    // By default make all the image styles for the consumer available, but
    // allow further refinement.
    if (
      !empty($configuration['styles']['refine']) &&
      !empty($configuration['styles']['custom_selection'])
    ) {
      $refined_style_ids = $configuration['styles']['custom_selection'];
      $image_style_ids = array_intersect($refined_style_ids, $image_style_ids);
    }
    return $image_style_ids;
  }

  /**
   * {@inheritdoc}
   */
  protected function doTransform($value, Context $context) {
    return $value;
  }

  /**
   * {@inheritdoc}
   */
  public function getOutputJsonSchema() {
    return [
      'type' => 'object',
      'properties' => [
        'data' => [
          'type' => 'object',
          'properties' => [
            'type' => ['type' => 'string'],
            'id' => ['type' => 'string'],
            'meta' => [
              'type' => 'object',
              'properties' => [
                'height' => [
                  'anyOf' => [
                    ['type' => 'integer'],
                    ['type' => 'null'],
                    ['type' => 'string'],
                  ],
                ],
                'width' => [
                  'anyOf' => [
                    ['type' => 'integer'],
                    ['type' => 'null'],
                    ['type' => 'string'],
                  ],
                ],
                'alt' => [
                  'anyOf' => [
                    ['type' => 'string'],
                    ['type' => 'null'],
                  ],
                ],
                'title' => [
                  'anyOf' => [
                    ['type' => 'string'],
                    ['type' => 'null'],
                  ],
                ],
                'uri' => ['type' => 'string'],
                'url' => ['type' => 'string', 'format' => 'uri']
              ],
            ],
          ],
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getSettingsForm(array $resource_field_info) {
    $settings = empty($resource_field_info['enhancer']['settings'])
      ? $this->getConfiguration()
      : $resource_field_info['enhancer']['settings'];
    return [
      'absolute' => [
        '#type' => 'checkbox',
        '#element_validate' => [[static::class, 'stringToBool']],
        '#title' => $this->t('Absolute URL'),
        '#description' => $this->t('Format as web accessible absolute URL, otherwise relative to the site root.'),
        '#default_value' => empty($settings['absolute']) ? FALSE : $settings['absolute'],
      ],
    ];
    return [];
  }

  /**
   * Transforms the string from the checkbox into an actual boolean.
   */
  public static function stringToBool(array &$element, FormStateInterface $form_state) {
    $new_value = (bool) $form_state->getValue($element['#parents']);
    $form_state->setValue($element['#parents'], $new_value);
  }

  /**
   * Transforms the mapping into a sequence.
   */
  // public static function mappingToSequence(array &$element, FormStateInterface $form_state) {
  //   $new_value = array_keys(array_filter($form_state->getValue($element['#parents'])));
  //   $form_state->setValue($element['#parents'], $new_value);
  // }

}
