<?php
namespace Drupal\viz_builder;

use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Theme\ThemeManagerInterface;
use Drupal\Core\Url;
use Twig\Environment;
use Twig\Extension\AbstractExtension;
use Twig\Loader\ChainLoader;
use Twig\TwigFunction;

class VizBuilderTwigExtension extends AbstractExtension {
  /**
   * The module handler to invoke alter hooks.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The theme manager to invoke alter hooks.
   *
   * @var \Drupal\Core\Theme\ThemeManagerInterface
   */
  protected $themeManager;

  protected $urlGenerator;

  protected $filename;

  /**
   * Constructs the TwigTweakExtension object.
   */
  public function __construct(ModuleHandlerInterface $module_handler, ThemeManagerInterface $theme_manager) {
    $this->moduleHandler = $module_handler;
    $this->themeManager = $theme_manager;

    $this->sdk = \Drupal::service('viz_builder.sdk');
    /** @var \Drupal\Core\File\FileUrlGenerator */
    $this->urlGenerator = \Drupal::service('file_url_generator');
  }

  public function getName() {
    return 'viz_builder';
  }

  /**
   * {@inheritdoc}
   */
  public function getFunctions(): array {
    $context_options = ['needs_context' => TRUE];
    $all_options = ['needs_environment' => TRUE, 'needs_context' => TRUE];

    $functions = [
      new TwigFunction('debug', [self::class, 'debug'], $all_options),
      new TwigFunction('url', [self::class, 'url']),
      new TwigFunction('add_js', [self::class, 'addJs'], $all_options),
      new TwigFunction('add_css', [self::class, 'addCss'], $all_options),
      new TwigFunction('add_block_js', [self::class, 'addBlockJs'], $all_options),
      new TwigFunction('add_block_css', [self::class, 'addBlockCss'], $all_options),
      new TwigFunction('assets', [self::class, 'assets'], ['needs_environment' => TRUE]),
      new TwigFunction('get_block', [self::class, 'getBlock']),
      new TwigFunction('get_node', [self::class, 'getNode']),
      new TwigFunction('get_collection', [self::class, 'getCollection']),
      new TwigFunction('get_categories', [self::class, 'getCategories']),
    ];

//    $this->moduleHandler->alter('twig_tweak_functions', $functions);
//    $this->themeManager->alter('twig_tweak_functions', $functions);

    return $functions;
  }

  public static function assets(Environment $environment, $asset_uri, $absolute = FALSE) {
    /* get calling template source
    $trace = debug_backtrace(DEBUG_BACKTRACE_PROVIDE_OBJECT, 2)[0];
    if (($template = @$trace['object']) && get_parent_class($template) == 'Twig\Template') {
      dump([$asset_uri => $template->getSourceContext()]);
    }*/
    $src_file = static::findPath($environment, $asset_uri);
//    $version = filemtime(\Drupal::service('kernel')->getAppRoot() . '/' . $src_file);

    if ($absolute) {
      return ($src_file); // . '?v=' . $version;
    }
    else {
      return base_path() . $src_file; // . '?v=' . $version;
    }
  }

  public static function getNode($id, $view_mode = 'full') {
    return (\Drupal::service('viz_builder.sdk')->getNode($id, $view_mode);
  }

  public static function getCollection($options = []) {
    return (\Drupal::service('viz_builder.sdk')->getCollection($options);
  }

  public static function getCategories($options = []) {

  }

  public static function url($uri) {
    if (preg_match('/^https?:\/\//', $uri)) {
      return $uri;
    }
    $uri = ltrim($uri, '/');

    if (preg_match('/(node|taxonomy|qy_page)/i', $uri)) {
      $uri = 'internal:/' . $uri;
      $url = Url::fromUri($uri);
      return $url->toString();
    }

    return base_path() . $uri;
  }

  public static function addJs(Environment $environment, $context, $uri, $weight = 0, $region = 'top') {
    if (preg_match('/^https?:\/\//', $uri)) {
      $file_uri = $uri;
    }
    else {
      $namespace = NULL;
      if ($uri[0] == '@') {
        list($namespace, $uri) = explode('/', $uri, 2);
        $namespace = substr($namespace, 1);
      }
      $file_uri = static::findPath($environment, $uri, $namespace);
    }
    $key = $region == 'bottom' ? 'js_bottom' : 'js';
    static::addAsset($key, $file_uri, $weight);
  }

  public static function addCss(Environment $environment, $context, $uri, $weight = 0) {
    if (preg_match('/^https?:\/\//', $uri)) {
      $file_uri = $uri;
    }
    else {
      $namespace = NULL;
      if ($uri[0] == '@') {
        list($namespace, $uri) = explode('/', $uri, 2);
        $namespace = substr($namespace, 1);
      }
      $file_uri = static::findPath($environment, $uri, $namespace);
    }
    static::addAsset('css', $file_uri, $weight);
  }

  public static function addBlockJs(Environment $environment, $context, $uri, $weight = 0, $region = 'top') {
    if (isset($context['block_path'])) {
      if (preg_match('/^https?:\/\//', $uri)) {
        $file_uri = $uri;
      }
      else {
        $dir_root = dirname(substr($context['block_path'], strlen(\Drupal::root()) + 1));
        $file_uri = "$dir_root/$uri";
      }
      $key = $region == 'bottom' ? 'js_bottom' : 'js';
      static::addAsset($key, $file_uri, $weight, 'block');
    }
  }

  public static function addBlockCss(Environment $environment, $context, $uri, $weight = 0) {
    if (isset($context['block_path'])) {
      if (preg_match('/^https?:\/\//', $uri)) {
        $file_uri = $uri;
      }
      else {
        $dir_root = dirname(substr($context['block_path'], strlen(\Drupal::root()) + 1));
        $file_uri = "$dir_root/$uri";
      }
      static::addAsset('css', $file_uri, $weight, 'block');
    }
  }

  public static function addAsset($type, $uri, $weight, $scope = 'theme') {
    $store = &drupal_static(QY_BUILDER_MEMORY_STORE_KEY, []);

    if (preg_match('/^https?:\/\//', $uri)) {
      $url = $uri;
    }
    else {
      $url = \Drupal::service('file_url_generator')->generateAbsoluteString($uri);
      $url .= '?v=' . filemtime(\Drupal::service('kernel')->getAppRoot() . '/' . $uri);
    }

    $store[$type][$uri] = [$url, $weight, basename($uri), $scope];
  }

  public static function debug($env, $context, $vars = NULL) {
    $var_dumper = '\Symfony\Component\VarDumper\VarDumper';
    if (class_exists($var_dumper)) {
      call_user_func($var_dumper . '::dump', $vars ?? $context);
    }
  }

  protected static function findPath(Environment $env, $relative_uri, $namespace = NULL) {
    /** @var \Drupal\Core\Template\Loader\FilesystemLoader $loader */
    $loader = $env->getLoader();
    $root = \Drupal::root();
    try {
      $paths = [];
      if ($loader instanceof ChainLoader) {
        $theme = \Drupal::service('theme.manager')->getActiveTheme();
        if (!$namespace) {
          $namespace = $theme->getName();
        }
        $loaders = $loader->getLoaders();
        foreach ($loaders as $_loader) {
          if ($_loader instanceof \Twig\Loader\FilesystemLoader) {
            $paths = array_flip($_loader->getPaths()) + array_flip($_loader->getPaths($namespace));
            $paths = array_keys($paths);
            break;
          }
        }
      }
      else {
        if (!$namespace) {
          $namespace = $loader::MAIN_NAMESPACE;
        }
        $paths = $loader->getPaths($namespace);
      }

      foreach ($paths as $path) {
        $theme_root = dirname(ltrim(str_replace($root, '', $path), '/'));
        $src_file = "$theme_root/$relative_uri";
        if (file_exists($src_file)) {
          return $src_file;
        }
      }
    } catch (\Throwable $e) {
    }
    return '';
  }
}
