<?php
namespace Drupal\viz_builder;

use Drupal\Component\Utility\Xss;
use Drupal\Core\Database\Database;
use Drupal\Core\Menu\MenuTreeParameters;
use Drupal\Core\Site\Settings;
use Drupal\Core\Url;
use Drupal\node\Entity\Node;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\node\NodeInterface;
use Drupal\user\UserInterface;
use Drupal\taxonomy\TermInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\HttpException;

class SDK {

  /**
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $nodeStorage;

  /**
   * @var \Drupal\taxonomy\TermStorage;
   */
  protected $termStorage;



  protected $entityRepository;

  /**
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  protected $pagerInfo = null;

  public function __construct() {
    $this->nodeStorage = \Drupal::entityTypeManager()->getStorage('node');
    $this->termStorage = \Drupal::entityTypeManager()->getStorage('taxonomy_term');
    $this->entityRepository = \Drupal::service('entity.repository');
    $this->moduleHandler = \Drupal::service('module_handler');
  }

  protected function getQueryParams() {
    $request = \Drupal::request();
    $params = $request->query->all();
    $list = [];
    foreach ($params as $key => $val) {
      $list[$key] = Xss::filter($val);
    }
    return $list;
  }

  protected function getPath() {
    return \Drupal::service('path.current')->getPath(\Drupal::request());
  }

  protected function getPageTitle($path = NULL) {
    if ($path) {
      $request = Request::create($path);
      $match = \Drupal::service('router')->matchRequest($request);
      $route = $match['_route_object'];
    }
    else {
      $request = \Drupal::request();
      $route = \Drupal::routeMatch()->getRouteObject();
    }

    $title = \Drupal::service('title_resolver')->getTitle($request, $route);
    if (is_array($title)) {
      return \Drupal::service('renderer')->renderPlain($title);
    }
    return $title;
  }

  protected function getDrupalPath($alias, $langcode = NULL) {
    /** @var \Drupal\path_alias\AliasManager $alias_manager */
    $alias_manager = \Drupal::service('path_alias.manager');
    $path = $alias_manager->getPathByAlias($alias, $langcode);

    return $path;
  }

  protected function getRouteParams () {
    $route_match = \Drupal::routeMatch();
    $params = $route_match->getParameters();
    $params_map = [];
    foreach ($params as $key => $param) {
      if ($param instanceof ContentEntityInterface) {
        $params_map[$key] = $this->serialize($param);
      } else {
        $params_map[$key] = $param;
      }
    }
    return $params_map;
  }

  protected function getDestination() {
    return \Drupal::destination()->get();
  }

  protected function getNode($id, $view_mode = 'full') {
    $node = $this->nodeStorage->load($id);
    $data = $this->serialize($node, $view_mode);
    $data['url'] = Url::fromUserInput("/node/$id")->toString();
    return $data;
  }

  protected function getNodes($type, $conds = [], $page = 0, $size = 12, $view_mode = 'list', $return_pager_object = FALSE) {
    $query = $this->nodeStorage->getQuery();
    $type = explode(',', $type);//support multiple values
    $query->condition('type', $type, 'IN');
    $query->accessCheck(TRUE);

    if (isset($conds['sort'])) {
      $sorts = array_map('trim', explode(',', $conds['sort']));
      foreach (array_filter(array_unique($sorts)) as $sort_key) {
        $direction = substr($sort_key, 0, 1) == '-' ? 'DESC' : 'ASC';
        $sort_key = str_replace('-', '', $sort_key);
        $query->sort($sort_key, $direction);
      }
      unset($conds['sort']);
    }

    foreach ($conds as $key => $value) {
      if ($key == 'keyword') {
        if (!empty(trim($value))) {
          $or = $query->orConditionGroup();
          $keyword = '%'. \Drupal::database()->escapeLike($value) . '%';
          $or->condition('title', $keyword, 'LIKE');
          $or->condition('body', $keyword, 'LIKE');
          //$or->condition('model', $keyword, 'LIKE');
          $query->condition($or);
        }
      }
      else {
        if (is_array($value)) {
          $query->condition($key, $value, 'IN');
        } else {
          $query->condition($key, $value);
        }
      }
    }

    $total = (clone $query)->count()->execute();
    $total_page = (int) ceil($total / $size);
    $this->pagerInfo = [
      'count' => (int) $total,
      'total_page' => (int) $total_page,
      'current_page' => (int) $page,
      'page_size' => (int) $size
    ];

    $ids = $query
      ->range($page * $size, $size)
      ->execute();
    $nodes = $this->nodeStorage->loadMultiple($ids);


    $data = [];
    foreach ($nodes as $node) {
      $data[] = $this->serialize($node, $view_mode);
    }
    return $data;
  }

  protected function getSubTermNode ($type, $conds = [], $page = 0, $size = 12, $search_fields = ['title', 'body'], $view_mode = 'list') {
    $query = $this->nodeStorage->getQuery()
      ->condition('type', $type);
    $query->accessCheck(TRUE);

    foreach ($conds as $key => $value) {
      if ($key == 'keyword') {
        if (!empty(trim($value))) {
          $or = $query->orConditionGroup();
          $keyword = '%'. \Drupal::database()->escapeLike($value) . '%';
          foreach ($search_fields as $field) {
            $or->condition($field, $keyword, 'LIKE');
          }
          $query->condition($or);
        }
      } else if ($key == 'term') {
        $term = $this->termStorage->load($value['id']);
        $sub_query = \Drupal::database()->select('taxonomy_term_field_data', 'td');
        $sub_query->leftJoin('taxonomy_term__parent', 'p', 'td.tid = p.entity_id');
        $sub_query->fields('p', ['entity_id', 'parent_target_id']);
        $sub_query->condition('td.vid', $term->bundle());
        $all_ids = $sub_query->execute()->fetchAllKeyed();
        $getChildrenIds = function($parent = 0, $depth = FALSE) use ($all_ids, &$getChildrenIds) {
          $ids = [];
          foreach ($all_ids as $id => $pid) {
            if ($parent == $pid) {
              $ids[] = $id;
              $children = $getChildrenIds($id);
              $ids = array_merge($ids, $children);
            }
          }
          return $ids;
        };

        $ids = $getChildrenIds($value['id']);
        array_unshift($ids, $value['id']);
        $query->condition($value['key'], array_unique($ids), 'IN');
      } else if ($key == 'sort') {
        if (is_array($value)) {
          $field = $value['field'];
          $dest = $value['dest'] ?? 'DESC';
          $query->sort($field, $dest);
        } else {
          $sorts = array_map('trim', explode(',', $value));
          foreach (array_filter(array_unique($sorts)) as $sort_key) {
            $direction = substr($sort_key, 0, 1) == '-' ? 'DESC' : 'ASC';
            $sort_key = str_replace('-', '', $sort_key);
            $query->sort($sort_key, $direction);
          }
        }
      } else {
        $query->condition($key, $value);
      }
    }

    $total = (clone $query)->count()->execute();
    $total_page = (int) ceil($total / $size);
    $this->pagerInfo = [
      'count' => (int) $total,
      'total_page' => (int) $total_page,
      'current_page' => (int) $page,
      'page_size' => (int) $size
    ];

    $ids = $query
      ->range($page * $size, $size)
      ->execute();
    $nodes = $this->nodeStorage->loadMultiple($ids);

    $data = [];
    foreach ($nodes as $node) {
      $data[] = $this->serialize($node, $view_mode);
    }
    return $data;
  }

  protected function getTerm($id, $view_mode = 'full') {
    $term = $this->termStorage->load($id);
    $data = $this->serialize($term, $view_mode);
    return $data;
  }

  /**
   * get terms by query
   * @param $conds Array
   *        [ gid => gid, module => module ]
   *
   */
  protected function getTerms($type, $conds = [], $page = 0, $size = 12, $view_mode = 'list') {
    $query = $this->termStorage->getQuery()
      ->condition('vid', $type);
    $query->accessCheck(TRUE);

    if (isset($conds['sort'])) {
      $sorts = array_map('trim', explode(',', $conds['sort']));
      foreach (array_filter(array_unique($sorts)) as $sort_key) {
        $direction = substr($sort_key, 0, 1) == '-' ? 'DESC' : 'ASC';
        $sort_key = str_replace('-', '', $sort_key);
        $query->sort($sort_key, $direction);
      }
      unset($conds['sort']);
    }
    else {
      $query->sort('weight', 'DESC');
    }

    foreach ($conds as $key => $value) {
      $query->condition($key, $value);
    }

    $ids = $query
      ->range($page * $size, $size)
      ->execute();
    $terms = $this->termStorage->loadMultiple($ids);

    $data = [];
    foreach ($terms as $term) {
      $data[] = $this->serialize($term, $view_mode);
    }
    return $data;
  }

  /**
   * Get terms by parent
   * @param $pid parent id
   * @param $vid vocabulary id
   * @param $max_depth int, depth of terms
   * @param $load_entities, bool TRUE|FALSE
   * @param $module article, media, vr, the name of content_type
   * @param $gid group id
   */
  protected function getTermTree($pid = 0, $vid = NULL, $max_depth = NULL, $load_entities = TRUE, $module = NULL, $gid = NULL) {
    $root = $this->termStorage->load($pid);
    $vid = $vid ?? ($root ? $root->bundle() : 'category');
    $terms = $this->termStorage->loadTree($vid, $pid, NULL, $load_entities);
    $flatten = [];

    if ($root) {
      $flatten[] = $this->serialize($root) + ['pid' => 0];
    }

    foreach ($terms as $term) {
      $flatten[] = $this->serialize($term) + [
        'pid' => $term->parent->target_id,
        'depth' => $term->depth
      ];
    }
    $tree = static::buildTree($flatten);

    if ($root) {
      $first = reset($tree);
      $result = $first['children'] ?? [];
    }
    else {
      $result = $tree;
    }

    return $result;
  }

  /**
   * get all parents of this tid, which includes current term
   */
  protected function getTermTreeByChild($tid) {
    $data = [];
    $terms = $this->termStorage->loadAllParents($tid);
    foreach ($terms as $term) {
      $data[] = $this->serialize($term);
    }

    return $data;
  }

  /**
   * @return current group
   */
  protected function getGroup($gid = NULL) {

    $gid = $gid ?? $this->getGroupId();

    /** @var Group $group */
    $group = \Drupal::entityTypeManager()->getStorage('group')->load($gid);

    $group = $this->serialize($group, 'full');
    return $group;
  }

  protected function getGroupId() {
    return \Drupal::service('forum.handle')->getGId();
  }

  protected function getGroupMember() {
    $data = [];
    if (\Drupal::moduleHandler()->moduleExists('fd')) {
      $member = \Drupal\fd\Util\AppContextNegotiator::getMember();
      if ($member) {
        // /** @var \Drupal\user\Entity\User */
        // $user = $member->getEntity();
        // $data['uid'] = $user->id();
        // $data['member_id'] = $member->id();
        // $data['name'] = @$member->field_member_name->value ?: $user->getDisplayName();
        // $data['intro'] = @$member->field_member_intro->value;
        // $data['verification'] = @$member->verification->value;
        // $data['code'] = @$member->code->value;
        // $data['phone'] = @$member->phone->value;
        // $data['id_number'] = @$member->id_number->value;
        $data = \Drupal::service('viz_builder.serializer')->processMemberEntity($member);
      }
    }
    return $data;
  }

  protected function getCollection($type, $options = []) {
    $params = \Drupal::request()->query;
    $bundle = $params->get('type', $type ?? FALSE);
    $gid    = $params->get('gid', $options['gid'] ?? FALSE);
    $limit  = $params->get('page_size', $options['page_size'] ?? 12);
    $page = $params->get('page', $options['page'] ?? 0);
    $cate_id = $params->get('cate_id', $options['cate_id'] ?? FALSE);
    $sub_cate_id = $params->get('sub_cate_id', $options['sub_cate_id'] ?? FALSE);
    $relationship = $params->get('relationship', $options['relationship'] ?? FALSE);
    $sorts = $params->get('sort', $options['sort'] ?? []);
    $langcode = $params->get('lang_code', $options['lang_code'] ?? FALSE);
    $q_tags = $params->get('tags', $options['tags'] ?? FALSE);
    $search = $params->get('search', $options['search'] ?? FALSE);
    $status = $params->get('status', $options['status'] ?? NULL);
    $sticky = $params->get('sticky', $options['sticky'] ?? FALSE);

    if (!$gid) {
      //      throw new BadRequestHttpException('missing gid');
    }

    $results = [];
    /** @var Group $group */
    $group = \Drupal::entityTypeManager()->getStorage('group')->load($gid);
    //    \Drupal::database()->query("SET SESSION sql_mode = ''")->execute();
    /** @var \Drupal\Core\Database\Query\Select $query */
    $query = \Drupal::database()
      ->select('group_content_field_data', 'gnode');
      // ->extend('Drupal\fd\Database\Query\FdPagerSelectExtender');
    $query->limit($limit);
    $query->leftJoin('node_field_data', 'n', 'gnode.entity_id = n.nid');
    $query->addTag('node_access');
    $query->addField('gnode', 'entity_id');
    $query->condition('gnode.gid', $gid);
    $query->addField('n', 'sticky');
    $query->orderBy('n.sticky', 'DESC');

    //    $sorts = explode(',', $sort_string);
    foreach (array_filter(array_unique($sorts)) as $sort_key) {
      $direction = substr($sort_key, 0, 1) == '-' ? 'DESC' : 'ASC';
      $sort_key = str_replace('-', '', $sort_key);
      $query->addField('n', $sort_key);
      $query->orderBy("n.$sort_key", $direction);
    }

    if ($langcode) {
      $query->condition('n.langcode', $langcode);
    }

    if ($cate_id) {
      if ($sub_cate_id) {
    //        $sub_cate_id = explode('-', $sub_cate_id);
        /**
         * 多个子分类的情况 subCateId=12-13-14
         *
         * relationship=1 :
         *    交集, 即分类cid=1且cid=2才会显示
         * default:
         *    并集, 即 cid IN (1,2,3)
         *
         * 如果更复杂，需要 (并集) && (并集），则传参数：
         *
         * 12-13-15:16 => 12 && 13 && (15 OR 16)
         */
        if ($relationship) {
          $n=1;
          foreach ($sub_cate_id as $sid) {
            $n++;
            if (is_numeric($sid)) {
              $query->innerJoin("node__field_category", "cate_$n", 'gnode.entity_id = %alias.entity_id AND %alias.field_category_target_id =  ' . $sid);
            }
            else {
              $sids = explode(":", $sid);
              $query->innerJoin("node__field_category", "cate_$n", 'gnode.entity_id = %alias.entity_id AND %alias.field_category_target_id IN ( ' . join(",",$sids) . " )");
            }
          }
        }
        else {
          $query->leftJoin('node__field_category', 'cate_node', 'gnode.entity_id = cate_node.entity_id');
          $query->condition('cate_node.field_category_target_id', $sub_cate_id, 'IN');
        }
      }
      else {
        //        $all_ids = \Drupal::database()->query('SELECT cate_id, pid FROM fd_category WHERE group_id=:gid', [':gid' => $gid])->fetchAllKeyed();
        $sub_query = \Drupal::database()->select('taxonomy_term_field_data', 'td');
        $sub_query->leftJoin('taxonomy_term__parent', 'p', 'td.tid = p.entity_id');
        $sub_query->condition('td.gid', $gid);
        $sub_query->fields('p', ['entity_id', 'parent_target_id']);
        $all_ids = $sub_query->execute()->fetchAllKeyed();
        $getChildrenIds = function($parent = 0, $depth = FALSE) use ($all_ids, &$getChildrenIds) {
          $ids = [];
          foreach ($all_ids as $id => $pid) {
            if ($parent == $pid) {
              $ids[] = $id;
              $children = $getChildrenIds($id);
              $ids = array_merge($ids, $children);
            }
          }
          return $ids;
        };

        $ids = $getChildrenIds($cate_id);
        array_unshift($ids, $cate_id);
        $ids = array_unique($ids);

        $query->leftJoin('node__field_category', 'cate_node', 'gnode.entity_id = cate_node.entity_id');
        $query->condition('cate_node.field_category_target_id', $ids, 'IN');
      }
    }

    //query by tags

    if (!empty($q_tags)) {
      //      $q_tags = explode('-', $q_tags);
      $q_tags = array_filter($q_tags, function($value){
        return is_numeric($value);
      });
      $query->leftJoin('node__field_tags', 'tags', 'n.nid = tags.entity_id');
      $query->condition('tags.field_tags_target_id', $q_tags, 'IN');
    }

    if ($status !== NULL) {
      $query->condition('n.status', (bool) $status);
    }

    if ($sticky) {
      $query->condition('n.sticky', 1);
    }

    if (!empty($search)) {
      $query->condition('n.title', '%'. $query->escapeLike($search) . '%', 'LIKE');
    }

    if ($bundle) {
      $query->condition('gnode.type', $group->bundle() . '-group_node-' . $bundle);
    }

    $query->distinct();
    $total = $query->countQuery()->execute()->fetchField();

    $total_page = (int)ceil($total / $limit);
    $this->pagerInfo = [
      'count' => (int) $total,
      'total_page' => (int) $total_page,
      'current_page' => (int) $page,
      'page_size' => (int) $limit
    ];

    $ids = $query->execute()->fetchCol();
    /** @var Node[] $nodes */
    $nodes = $this->nodeStorage->loadMultiple($ids);
    foreach ($nodes as $node) {
      $node = $this->entityRepository->getTranslationFromContext($node, $langcode);
      $item = $this->serialize($node, 'list');
      //      unset($item['body']);
      //      unset($item['access_code']);
      $item['url'] = Url::fromUserInput("/node/{$node->id()}")->toString();
      $results[] = $item;
    }

    return $results;
  }

  protected function getRelatedNodes($nid, $limit = 8) {
    if (!$this->moduleHandler->moduleExists('fd')) {
      throw new HttpException('Feature recommendation not enabled.');
    }
    $controller = \Drupal\fd\Controller\RecommendApiController::create(\Drupal::getContainer());
    $result = $controller->recommend($limit, $nid);
    return $result;
  }

  protected function getSearchResult() {
    if (!$this->moduleHandler->moduleExists('fd_search')) {
      throw new HttpException(400, 'Feature search not enabled');
    }

    try {
      $request = \Drupal::request();
      $controller = \Drupal\fd_search\Controller\FdSearchController::create(\Drupal::getContainer());
      $response = $controller->searchAll($request);
      $content = $response->getContent();

      $data = json_decode($content, TRUE);

      $this->pagerInfo = [
        'count' => (int) $data['total'],
        'total_page' => (int) ceil($data['total'] / $data['page_size']),
        'current_page' => (int) $data['page'],
        'page_size' => (int) $data['page_size']
      ];

      return $data;
    }
    catch (\Exception $e) {
      return [
        'list' => []
      ];
    }
  }

  protected function getEntity($entity_type_id, $entity_id) {
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);
    $entity = $storage->load($entity_id);
    $data = $this->serialize($entity);
    return $data;
  }

  protected function getEntities($entity_type_id, $opts = [], $view_mode = 'full') {
    $conds = $opts['conds'] ?? [];
    $page = $opts['page'] ?? 0;
    $size = $opts['page_size'] ?? 20;
    $storage = \Drupal::entityTypeManager()->getStorage($entity_type_id);

    $query = $storage->getQuery();
    $query->accessCheck(TRUE);
    if (isset($opts['sort'])) {
      $sorts = array_map('trim', explode(',', $opts['sort']));
      foreach (array_filter(array_unique($sorts)) as $sort_key) {
        $direction = substr($sort_key, 0, 1) == '-' ? 'DESC' : 'ASC';
        $sort_key = str_replace('-', '', $sort_key);
        $query->sort($sort_key, $direction);
      }
    }

    foreach ($conds as $key => $cond) {
      switch ($key) {
        case 'keyword':
          $entity_type = $storage->getEntityType();
          $keyword = '%'. \Drupal::database()->escapeLike($cond) . '%';
          $query->condition($entity_type->getKey('label'), $keyword, 'LIKE');
          break;
        default:
          $query->condition($key, $cond, is_array($cond) ? 'IN' : '=');
      }
    }

    $total = (clone $query)->count()->execute();
    $total_page = (int) ceil($total / $size);
    $this->pagerInfo = [
      'count' => (int) $total,
      'total_page' => (int) $total_page,
      'current_page' => (int) $page,
      'page_size' => (int) $size
    ];

    $ids = $query
      ->range($page * $size, $size)
      ->execute();
    $nodes = $storage->loadMultiple($ids);

    $data = [];
    foreach ($nodes as $node) {
      $data[] = $this->serialize($node, $view_mode);
    }

    return $data;
  }

  protected function getDynamic($func_name, $args = []) {
    if (is_callable($func_name)) {
      return call_user_func_array($func_name, $args);
    }
    throw new HttpException(400, 'No matched results.');
  }

  public function serialize($entity, $view_mode = 'full') {
    $data = \Drupal::service('viz_builder.serializer')->processEntity($entity);
    \Drupal::moduleHandler()->alter('entity_serialize', $data, $entity, $view_mode);
    return (array) $data;
  }

  /**
   * SDK的方法执行前，切换到API数据库，执行后切换回原数据库
   */
  public function __call($name, $arguments) {
    if (substr($name, 0, 3) == 'get') {
      try {
        $reflection = new \ReflectionClass($this);
        $method = $reflection->getMethod($name);
        if ($method->isProtected()) {
          //$db_key = Settings::get('qy_builder.sdk.database_key', 'default');
          //$origin_key = Database::setActiveConnection($db_key);
          $result = call_user_func_array([$this, "$name"], $arguments);
          //Database::setActiveConnection($origin_key);
          return $result;
        }
      }
      catch (\ReflectionException $e) {
        return 'Method not exists';
      }
    }
  }

  public function getPagerInfo () {
    if (!empty($this->pagerInfo)) {
      return $this->pagerInfo;
    }
    return [];
  }

  /**
   * get Current User
   */
  public function getCurrentUser() {
    $current_user = \Drupal::currentUser();
    return $current_user;
  }


  public function getLoginFormBuildId () {
    $formBuilder = \Drupal::service('form_builder');
    $form = $formBuilder->getForm('\Drupal\user\Form\UserLoginForm');
    $formBuildId = $form['#build_id'];
    return $formBuildId;
  }

  public function getMenuLinkList ($menu_name, $depth = 1) {
    if (!$menu_name) {
      return null;
    }
    $menu_tree =$this->doGetMenu($menu_name, $depth);
    $menu_list = $this->menuLinkList($menu_tree['#items']);
    return $menu_list;
  }

  public function getMenuLinkTree ($menu_name, $depth = 1) {
    if (!$menu_name) {
      return null;
    }
    $menu_tree =$this->doGetMenu($menu_name, $depth);
    $menu_list = $this->menuLinkTree($menu_tree['#items']);
    return $menu_list;
  }

  protected function doGetMenu ($menu_name, $depth) {
    $handle = \Drupal::menuTree();
    $params = $handle->getCurrentRouteMenuTreeParameters($menu_name);
    $params->setMinDepth(1)->setMaxDepth($depth);
    $menu = $handle->load($menu_name, $params);

    $manipulators = [
      ['callable' => 'menu.default_tree_manipulators:checkNodeAccess'],
      ['callable' => 'menu.default_tree_manipulators:checkAccess'],
      ['callable' => 'menu.default_tree_manipulators:generateIndexAndSort']
    ];

    $tree = $handle->transform($menu, $manipulators);
    $menu = $handle->build($tree);

    $items = $menu['#items'];
    if ($items) {
      $this->findLastActiveTrail($items, []);
      $menu['#items'] = $items;
    }

    return $menu;
  }

  protected function findLastActiveTrail (&$items) {
    $return = True;
    foreach ($items as &$item) {
      if($item['in_active_trail']) {;
        if ($item['below']) {
          if ($this->findLastActiveTrail($item['below'])) {
            $item['is_last_active_trail'] = true;
          }
        } else {
          $item['is_last_active_trail'] = true;
	}
        return False;
      }
    }
    return $return;
  }

  protected function menuLinkTree ($items, $depth = 1) {
    $menu_links = [];
    foreach ($items as $item) {
      $tmp = [
        'title' => $item['title'],
        'url'   => $item['url']->toString(),
        'depth' => $depth,
        'options' => $item['url']->getOptions(),
        'is_expanded' => $item['is_expanded'],
        'is_collapsed' => $item['is_collapsed'],
        'in_active_trail' => $item['in_active_trail'],
      ];
      if (isset($item['is_last_active_trail'])) {
        $tmp['is_last_active_trail'] = $item['is_last_active_trail'];
      }
      if (isset($tmp['options']['menu_icon']) && isset($tmp['options']['menu_icon']['fid'])) {
         $fid  = $tmp['options']['menu_icon']['fid'] ?: 0;
         $file =  \Drupal\file\Entity\File::load($fid);
         if ($fid && $file) $tmp['options']['menu_icon']['url'] = $file->createFileUrl();
      }
      if ($item['below']) {
        $tmp['children'] = $this->menuLinkTree($item['below'], $depth + 1);
      }
      $menu_links[] = $tmp;
    }
    return $menu_links;
  }

  protected function menuLinkList ($items, $parent = '', $depth = 1) {
    $list = [];
    $i = 1;
    foreach ($items as $item) {
      $tmp = [
        'title' => $item['title'],
        'url'   => $item['url']->toString(),
        'depth' => $depth
      ];
      $uuid = $item['original_link']->getDerivativeId();
      $tmp['parent'] = $parent;
      $tmp['uuid'] = $uuid;
      if ($item['below']) {
        $childList = $this->menuLinkList($item['below'], $uuid, $depth + 1);
      }
      unset($tmp['children']);
      $list [] = $tmp;
      array_push($list, ...$childList);
      $i++;
    }
    return $list;
  }

  //get node view statistics
  //return [total_count, day_count, timestamp]
  //https://api.drupal.org/api/drupal/core%21modules%21statistics%21src%21StatisticsViewsResult.php/class/StatisticsViewsResult/9
  public function getNodeStatistics($nid) {
    if ( \Drupal::hasService('statistics.storage.node')
	&&  $stat = \Drupal::service('statistics.storage.node')->fetchView($nid)) {
      return [  $stat->getTotalCount(),  $stat->getDayCount(),  $stat->getTimestamp() ];
    }
    else {
      return [0,0,0];
    }

  }

  public function getForumCategoryList () {
    return \Drupal::service('forum.handle')->getCategoryList();
  }

  public function getForumStickyPostList ($options = []) {
    return \Drupal::service('forum.handle')->getStickyPostList($options);
  }

  public function getForumHotPostList ($options = []) {
    return \Drupal::service('forum.handle')->getHotPostList($options);
  }

  public function getForumPostList ($options = []) {
    return \Drupal::service('forum.handle')->getPostList($options);
  }

  public function getForumSearchList ($options = []) {
    return \Drupal::service('forum.handle')->getSearchList($options);
  }

  public function getForumUserPostList ($options = []) {
    return \Drupal::service('forum.handle')->getUserPost($options);
  }

  public function getForumPostContent ($options = []) {
    return \Drupal::service('forum.handle')->getPostContent($options);
  }

  public function getForumReplyList ($options = []) {
    return \Drupal::service('forum.handle')->getReplyList($options);
  }

  public function getForumTagList () {
    return \Drupal::service('forum.handle')->getForumTagList();
  }

  public function getForumPostSortInfo ($options = []) {
    return \Drupal::service('forum.handle')->getPostSortInfo($options);
  }

  public function getBlockContentIdFromSetting ($key) {
    $res = \Drupal::entityTypeManager()->getStorage('content_setting')->loadByProperties([
      'key' => $key
    ]);
    if ($res) {
      $entity = current($res);
      return $entity->get('setting')->value;
    }
    return false;
  }

  public function getDrupalEntity($entity_type, $id = NULL, $view_mode = 'full', $langcode = NULL, $check_access = TRUE) {
    $tweakHandle = \Drupal::service('twig_tweak.twig_extension');
    $build = $tweakHandle->drupalEntity($entity_type, $id, $view_mode, $langcode, $check_access);
    return \Drupal::service('renderer')->renderPlain($build);
  }

  public function getDrupalBlock($id, array $configuration = [], $wrapper = TRUE) {
    $tweakHandle = \Drupal::service('twig_tweak.twig_extension');
    $build = $tweakHandle->drupalBlock($id, $configuration, $wrapper);
    return \Drupal::service('renderer')->renderPlain($build);
  }

  /**
   * Copy from fd.utility
   * @param $flatten
   * @param string $pidKey
   * @param string $idKey
   *
   * @return mixed
   */
  public static function buildTree($flatten, $pidKey = 'pid', $idKey = 'id') {
    $parents = array();

    foreach ($flatten as $item){
      $parents[$item[$pidKey]][] = $item;
    }

    $fnBuilder = function($items, $parents, $idKey) use (&$fnBuilder) {

      foreach ($items as $position => $item) {

        $id = $item[$idKey];

        if(isset($parents[$id])) { //is the parent set
          $item['children'] = $fnBuilder($parents[$id], $parents, $idKey); //add children
        }

        //reset the value as children have changed
        $items[$position] = $item;
      }

      //return the item
      return $items;
    };

    $parent = !empty($parents[0]) ? $parents[0] : [];

    return $fnBuilder($parent, $parents, $idKey);
  }
}
