<?php
namespace Drupal\viz_builder;

use Drupal\filter\FilterProcessResult;
use Drupal\filter\Render\FilteredMarkup;
use Drupal\group\Entity\GroupMembershipInterface;


/**
 * @file
 * 实现了entity转换成data数组的所有方法
 * 可以直接序列化大多数字段，特殊字段通过hook_entity_serialize_alter来进行实现或者修改
 */

class Serializer {
  /**
   * implements hook_entity_serialize
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   * @return array
   */
  function processEntity(\Drupal\Core\Entity\ContentEntityInterface $entity) {
    if (is_null($entity)) {
      return [];
    }

    $data = [
      'id'    => $entity->id(),
      'uuid'  => $entity->uuid(),
      'url'   => $entity->toUrl()->toString(),
      'title' => $entity->label(),
      'label' => $entity->label(),
      'entity_type'   => $entity->getEntityTypeId(),
      'entity_bundle' => $entity->bundle(),
    ];

    switch ($data['entity_type']) {
      case 'node':
        /** @var \Drupal\node\NodeInterface $entity */
        $data['user_name'] = $entity->getOwner()->getDisplayName();
        $data['created'] = $entity->getCreatedTime();
        $data['changed'] = $entity->get('changed')->value;

        $data['createdat'] = \Drupal::service('date.formatter')->format($entity->getCreatedTime(), 'custom', DATE_RFC3339);
        $data['changedat'] = \Drupal::service('date.formatter')->format($entity->get('changed')->value, 'custom', DATE_RFC3339);


        if ($entity->hasField('body')) {
          $data['formatted'] = NULL;
          $data['summary'] = $entity->get('body')->summary;
          $text = $entity->get('body')->value;
          if (!empty($text)) {
            $data['formatted'] = $this->processFormatText($text, $entity->get('body')->format);
            if (empty($data['summary'])) {
              $summary = strip_tags(text_summary($data['formatted'], $entity->get('body')->format, 120));
            }
            $data['summary'] = $summary;
          }
        }
        break;

      case 'taxonomy_term':
        /** @var \Drupal\taxonomy\TermInterface $entity */
        $data['desc'] = $entity->getDescription();
        break;

      case 'lecture_unit':
        $data['cover'] = NULL;
        $data['media'] = NULL;

        $length = $entity->media_length->value;
        $data['media_length_raw'] = $length;
        $data['media_length'] = gmdate($length < 3600 ? 'i:s' : 'H:i:s', $length);

        if ($file = $entity->cover->entity) {
          $data['cover'] = $this->processFile($file);
        }

        if ($file = $entity->file->entity) {
          $data['media'] = $this->processFile($file);
        }
        break;

      case 'group':
      default:
        break;

    }

    /** 遍历 Field API 字段 */
    foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
      if (!$field_definition instanceof \Drupal\field\FieldConfigInterface) { //only for field_ui attached fields
        continue;
      }
      if (isset($data[$field_name])) {  //skip when processed
        continue;
      }
      $this->_processFields($data, $entity, $field_name, $field_definition);
    };
    return $data;
  }

  /**
   * @description 引用字段
   * @param $field_name string
   * @param $entity \Drupal\Core\Entity\ContentEntityInterface
   *
   * 不会渲染普通的entity，只渲染特殊的entity，比如term、image、paragraph、文件等。
   * 普通的entity如果继续渲染，就会行成死循环
   */
  function processReferencedField($entity, $field_name)
  {
    $field_definition = $entity->getFieldDefinition($field_name);

    if ($field_definition->getType() == 'file'
      || $field_definition->getType() == 'image'
    ) {
      $callback = 'processFile';
    }

    $paragraph_exist = \Drupal::moduleHandler()->moduleExists('paragraphs');
    if ($field_definition->getType() == 'entity_reference_revisions' && $paragraph_exist) {
      $callback = 'processParagraph';
    }

    $target_type = $field_definition->getFieldStorageDefinition()->getSetting('target_type');

    if ($target_type == 'taxonomy_term' || $target_type == 'media') {
      $callback = 'processTargetEntity';
    }

    if ($target_type == 'node') {
      $callback = 'processEntity';
    }

    if (isset($callback)) {
      $origin_entity = $entity;
      foreach ($origin_entity->get($field_name)->referencedEntities() as $key => $entity) {
        $value = call_user_func([$this, $callback], $entity);
        //fix image field alt value
        if ($field_definition->getType() === 'image') {
          if ($value) {
            $origin_array = $origin_entity->get($field_name)->getValue();
            $origin_value = $origin_array[$key];
            if ($origin_value['alt']) {
              $value['alt'] = $origin_value['alt'];
            }
          }
        }
        $data[] = $value;
      }
    }
    return isset($data) ? $data : null;
  }



  /**
   * @description 文件字段
   * @param $file \Drupal\file\FileInterface
   * @return array
   */
  function processFile($file)
  {
    $data = [];
    if (!is_null($file)) {
      $data = [
        'fid' => $file->id(),
        'url' => $file->createFileUrl(FALSE),
        'alt' => @$file->alt->value ?? $file->getFilename(),
        'title' => $file->getFilename(),
      ];
    }
    return $data;
  }

  /**
   * @param $field \Drupal\link\Plugin\Field\FieldType\LinkItem | null
   * @return array
   */
  function processLink($field)
  {
    $data = [];
    if (!is_null($field) && !$field->isEmpty()) {
      $data = [
        'url' => $field->getUrl()->toString(),
        'uri' => $field->getUrl()->toUriString(),
        'title' => $field->get('title'),
      ];
    }
    return $data;
  }

  /**
   * 引用字段
   *
   * @param $term \Drupal\taxonomy\TermInterface
   *
   * process Media | Term
   * @return array
   */
  function processTargetEntity($entity)
  {
    if (is_null($entity)) {
      return NULL;
    }

    $data['id']    = $entity->id();
    $data['uuid']  = $entity->uuid();
    $data['label'] = $entity->label();
    $data['url']   = $entity->toUrl()->toString();
    foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
      if (!$field_definition instanceof \Drupal\field\FieldConfigInterface) { //only for field_ui attached fields
        continue;
      }
      if (isset($data[$field_name])) {  //skip when processed
        continue;
      }
      $this->_processFields($data, $entity, $field_name, $field_definition);
    };

    return $data;
  }



  /**
   * @description paragraph字段
   * @param $para_entity \Drupal\paragraphs\Entity\Paragraph
   * @return array
   */
  function processParagraph($para_entity)
  {
    if (is_null($para_entity)) {
      return NULL;
    }

    $data = [
      'paragraph_type' => $para_entity->bundle()
    ];

    foreach ($para_entity->getFieldDefinitions() as $field_name => $field_definition) {
      $this->_processFields($data, $para_entity, $field_name, $field_definition);
    }
    if (isset($data[$field_name][0]) && empty($data[$field_name][0])) {
      $data[$field_name] = null;
    }
    return $data;
  }

  /**
   * Format a text with formatter
   * @param $text string
   * @param $format string
   * @return \Drupal\Component\Render\MarkupInterface|string
   */
  function processFormatText($text, $format)
  {

    if (!isset($text) || $text === '') {
      $processed = new FilterProcessResult('');
    } else {
      $build = [
        '#type' => 'processed_text',
        '#text' => $text,
        '#format' => $format ?: filter_default_format(),
        '#filter_types_to_skip' => [],
        //'#langcode' => $entity->get('body')->getLangcode(),
      ];
      // Capture the cacheability metadata associated with the processed text.
      $processed_text = \Drupal::service('renderer')->renderPlain($build);
      $processed = FilterProcessResult::createFromRenderArray($build)
        ->setProcessedText((string)$processed_text);
    }
    return FilteredMarkup::create($processed->getProcessedText());
  }

  /**
   * 内部方法，参考processParagraph
   * @param $entity \Drupal\Core\Entity\Entity
   */
  private function _processFields(&$data, &$entity, $field_name, $field_definition)
  {
    $cardinality = $field_definition->getFieldStorageDefinition()->getCardinality();
    $count = $entity->get($field_name)->count();
    $field_type = $field_definition->getType();
    //基础字段
    if (in_array($field_type, ['text', 'string', 'integer', 'float', 'email', 'decimal', 'boolean'])) {
      for ($i = 0; $i < $count; $i++) {
        $value = $entity->get($field_name)->get($i)->value;
        if ($field_type == 'integer') {
          $value = (int) $value;
        }
        if ($field_type == 'boolean') {
          $value = (bool) $value;
        }
        $data[$field_name][] = $value;
      }
    }

    //特殊多值字段: link
    if ($field_type == 'link') {
      for ($i = 0; $i < $count; $i++) {
        $data[$field_name][] = $this->processLink($entity->get($field_name)->get($i));
      }
    }

    //特殊多值字段: long text
    if (in_array($field_type, ['text_long', 'string_long'])) {
      for ($i = 0; $i < $count; $i++) {
        $format = $entity->get($field_name)->get($i)->format;
        $text = $entity->get($field_name)->get($i)->value;
        $data[$field_name][] = $this->processFormatText($text, $format);
      }
    }

    //引用字段
    if (in_array($field_type, ['image', 'file', 'entity_reference', 'entity_reference_revisions'])) {
      $data[$field_name] = $this->processReferencedField($entity, $field_name);
    }

    //引用字段只会渲染term、image、paragraph、file，不会渲染默认entity，
    //这个是默认的entity，只生成一个url和id，需要渲染的话，要自定义代码处理
    if ($field_type == 'entity_reference' && !isset($data[$field_name]) && is_null($data[$field_name])) {
      foreach ($entity->get($field_name)->referencedEntities() as $ref_entity) {
        /** @var \Drupal\Core\Entity\EntityInterface $ref_entity */
        $data[$field_name][] = [
          'entity_type' => $ref_entity->getEntityTypeId(),
          'id' => $ref_entity->id(),
          'label' => $ref_entity->label(),
        ];
      }
    }

    if ($field_type == 'json_native' || $field_type == 'json') {
      for ($i = 0; $i < $count; $i++) {
        $data[$field_name][] = json_decode($entity->get($field_name)->get($i)->value, TRUE);
      }
    }

    //尝试获取默认的value
    if (!isset($data[$field_name])) {
      $values = $entity->get($field_name)->getValue();
      if (!is_null($values) && isset($values['value']) && !empty($values['value'])) {//only have value
        $data[$field_name][] = $values['value'];
      }
    }

    //cardinality
    if (isset($data[$field_name]) && $cardinality == 1) {
      $data[$field_name] = $data[$field_name][0];
    }
  }

  function processMemberEntity(GroupMembershipInterface $member) {
    /** @var User $user */
    $user = $member->getEntity();

    $user_data = [
      'id' => $user->id(),
      'name' => $user->getDisplayName(),
      'avatar' => null,
    ];

    if ($avatar = $user->user_picture->entity) {
      $user_data['avatar'] = \Drupal::service('file_url_generator')->generateAbsoluteString($avatar->getFileUri());
    }

    $data['user'] = $user_data;
    $data['id'] = $member->id();
    $data['name'] = @$member->field_member_name->value ?: $user_data['name'];
    $data['intro'] = @$member->field_member_intro->value;
    $data['verification'] = @$member->verification->value;
    $data['code'] = @$member->code->value;
    $data['phone'] = @$member->phone->value;
    $data['id_number'] = @$member->id_number->value;
    $data['tags'] = [];
    $data['avatar'] = $user_data['avatar'];
    $data['created'] = date('Y-m-d', $member->getCreatedTime());
    $data['role'] = array_column($member->group_roles->getValue(), 'target_id');
    $data['fee'] = @(float)$member->field_member_fee->value ?: 0;
    $data['weight'] = $member->field_weight->value ?: 0;

    if (is_object($member->get('category')) && $member->get('category')->getValue() ) {
      $data['category'] = array_column($member->get('category')->getValue(), 'target_id');
    } else {
      $data['category'] = [];
    }

    if ($file = @$member->field_member_avatar->entity) {
      $data['avatar'] = \Drupal::service('file_url_generator')->generateAbsoluteString($file->getFileUri());
    }

    if (isset($member->tas)) {
      foreach($member->tags as $item) {
        if ($term = $item->entity) {
          $data['tags'][] = [
            'id' => $term->id(),
            'uuid' => $term->uuid(),
            'name' => $term->label(),
          ];
        }
      }
    }

    //$membership = new GroupMembership($member);
    //$data['access_private_group'] = $membership->hasPermission('view private group');

    return $data;
  }

}
